/* driver.c   [ VMXON driver. ]
Does minimum necessary to do a VMXON. Builds upon our vmxmsr driver.
This driver uses a number of typedef/structs located in our driver.h
*/

#include <ntddk.h>
#include "driver.h" // PHYSICAL_ADDRESS, VMX_MSR, VIRT_CPU

/* our assembly prototypes (defined in the vmxmsr.asm in amd64 dir) */
ULONG64 _MyCpuid();
ULONG64 _ReadMsr(ULONG32 reg);
ULONG64 _Cr4();
VOID	_SetCr4(ULONG32 mask);
VOID	_VmxOn(PHYSICAL_ADDRESS PA);
ULONG32 _VmFailInvalid();

/* relevant MSRs and their indices */
#define MSR_IA32_VMX_BASIC   		0x480
#define MSR_IA32_FEATURE_CONTROL    0x03A

/* This is the bitmask to use for check/set of CR4 for VMX enable */
#define X86_CR4_VMXE				0x2000  // CR4.VMXE (bit 13) =

const char *legend[ 12 ] = {
	"IA32_VMX_BASIC_MSR",			// 0x480
	"IA32_VMX_PINBASED_CTLS_MSR",	// 0x481
	"IA32_VMX_PROCBASED_CTLS_MSR",	// 0x482
	"IA32_VMX_EXIT_CTLS_MSR",		// 0x483
	"IA32_VMX_ENTRY_CTLS_MSR",		// 0x484
	"IA32_VMX_MISC_MSR",			// 0x485
	"IA32_VMX_CR0_FIXED0_MSR",		// 0x486
	"IA32_VMX_CR0_FIXED1_MSR",		// 0x487
	"IA32_VMX_CR4_FIXED0_MSR",		// 0x488
	"IA32_VMX_CR4_FIXED1_MSR",		// 0x489
	"IA32_VMX_VMCS_ENUM_MSR"		// 0x48A
};

enum { legend_count = 11 };

/* Allocates contiguous physical memory, and zeroes it out
returns NULL if failure */
PVOID AllocateContiguousMemory(ULONG size)
{
	PVOID Address;
	PHYSICAL_ADDRESS l1, l2, l3;

	l1.QuadPart = 0;
	l2.QuadPart = -1;
	l3.QuadPart = 0x200000; /* don't know why (do you?), sorry */

	Address = MmAllocateContiguousMemorySpecifyCache(size, l1, l2, l3, MmCached);

	if (Address == NULL)
		return NULL;

	RtlZeroMemory(Address, size);
	return Address;
}

NTSTATUS SetupVMX(PVIRT_CPU pCpu)
{
	PHYSICAL_ADDRESS pa;
	ULONG64 msr;
	PVMX_BASIC_MSR pvmx;
	ULONG32 i;
	PVOID va;
	ULONG size;

	i = KeGetCurrentProcessorNumber();

	pCpu->ProcessorNumber = i;
	msr = _ReadMsr(MSR_IA32_VMX_BASIC);
	pvmx = (PVMX_BASIC_MSR)&msr;

	size = pvmx->szVmxOnRegion;

	DbgPrint("VMXON region size: 0x%x\n", size);
	DbgPrint("VMX revision ID: 0x%x\n", pvmx->RevId);

	va = AllocateContiguousMemory(size);

	if (va == NULL)
	{
		DbgPrint("error: can't allocate vmxon region\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	*(ULONG32 *)va = pvmx->RevId;
	pa = MmGetPhysicalAddress(va);

	_VmxOn(pa);

	if (_VmFailInvalid())
	{
		DbgPrint("_VmxOn failed\n");
		return STATUS_UNSUCCESSFUL;
	}

	pCpu->VMXON_va = va;
	pCpu->VMXON_pa = pa;

	va = AllocateContiguousMemory(size);

	if (va == NULL)
	{
		DbgPrint("error: can't allocate vmcs region\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	*(ULONG32 *)va = pvmx->RevId;        
	pa = MmGetPhysicalAddress(va);

	pCpu->VMCS_va = va;
	pCpu->VMCS_pa = pa;

	va = AllocateContiguousMemory(0x1000);
	pa = MmGetPhysicalAddress(va);

	if (va == NULL)
	{
		DbgPrint("error: can't allocate msr bitmap\n");
		return STATUS_INSUFFICIENT_RESOURCES;
	}

	pCpu->MSR_bitmap_va = va;
	pCpu->MSR_bitmap_pa = pa;

	return STATUS_SUCCESS;    
}

NTSTATUS CheckIfVMXIsEnabled()
{    
	ULONG64 cr4, msr;
	PCR4_REG pCr4;

	/* vmxon supported ? */
	_SetCr4(X86_CR4_VMXE);
	cr4 = _Cr4();
	pCr4 = (PCR4_REG) &cr4;

	if (!(cr4 & X86_CR4_VMXE))
	{
		DbgPrint("error: VMXON not supported\n");
		return STATUS_UNSUCCESSFUL;
	}

	DbgPrint("CR4: %#018llx ... CR4.VMXE=%d\n", cr4, pCr4->VMXE);

	/* vmx deactivated by BIOS? */        
	msr = _ReadMsr(MSR_IA32_FEATURE_CONTROL);
	if (!(msr & 4))
	{
		DbgPrint("vmx is disabled in bios: MSR_IA32_FEATURE_CONTROL is 0x%llx\n", msr);
		return STATUS_UNSUCCESSFUL;
	}

	DbgPrint("VMX appears supported (with IA32_FEATURE_CONTROL and CR4.VMXE)\n");
	return STATUS_SUCCESS;
}

void DriverUnload(PDRIVER_OBJECT pDriverObject)
{
	DbgPrint("Driver unloading\n");
}

NTSTATUS DriverEntry(PDRIVER_OBJECT DriverObject, PUNICODE_STRING RegistryPath) 
{
	ULONG64 ret;
	ULONG32 msr;
	char buff[16];
	const char** msrStr = legend;
	int ind = 0;
	VIRT_CPU vCpu;
	vCpu.Self = &vCpu;

	RtlZeroMemory(buff, sizeof(buff));
	DriverObject->DriverUnload = DriverUnload;
	ret = _MyCpuid();
	RtlCopyMemory(buff, (void*)ret, sizeof(buff));
	DbgPrint("ret :%p\n", ret);
	DbgPrint("Hello World! cpu:%s\n", buff);

	for(ind=0; ind < legend_count; ++ind) {
		msr = 0x480 + ind;
		DbgPrint("msr: %#010x  %30s	=> %#018llx\n", msr, *(msrStr++), _ReadMsr(msr));
	}

	CheckIfVMXIsEnabled();
	SetupVMX(&vCpu);

	return STATUS_SUCCESS;

}

